--[[
meta
	.comment
	.title
	.script_type
	.wrap_style
	.width
	.height
	.scaled_border_and_shadow
	.ratio
	.zoom
	.position
	.last_style
	.audio
	.video
styles
	[i]
		.name
		.fontname
		.fontsize
		.color1
		.alpha1
		.color2
		.alpha2
		.color3
		.alpha3
		.color4
		.alpha4
		.bold
		.italic
		.underline
		.strikeout
		.scale_x
		.scale_y
		.spacing
		.angle
		.borderstyle
		.outline
		.shadow
		.alignment
		.margin_l
		.margin_r
		.margin_v
		.encoding
lines
	[i]
		.comment
		.layer
		.start_time
		.end_time
		.duration
		.infade
		.outfade
		.style
		.styleref
		.actor
		.margin_l
		.margin_r
		.margin_v
		.effect
		.k_text
		.text
		.i
		.width
		.height
		.ascent
		.descent
		.intlead
		.extlead
		.x
		.y
		.left
		.center
		.right
		.top
		.middle
		.bottom
		.syls
			[i]
				.i
				.word_i
				.start_time
				.end_time
				.duration
				.text
				.inline_fx
				.prespace
				.postspace
				.width
				.height
				.ascent
				.descent
				.intlead
				.extlead
				.x
				.y
				.left
				.center
				.right
				.top
				.middle
				.bottom
		.chars
			[i]
				.i
				.syl_i
				.word_i
				.start_time
				.end_time
				.duration
				.text
				.width
				.height
				.ascent
				.descent
				.intlead
				.extlead
				.x
				.y
				.left
				.center
				.right
				.top
				.middle
				.bottom
		.words
			[i]
				.i
				.start_time
				.end_time
				.duration
				.text
				.width
				.height
				.ascent
				.descent
				.intlead
				.extlead
				.x
				.y
				.left
				.center
				.right
				.top
				.middle
				.bottom
]]--

--Global ASS tables
meta = {}
styles = {}
lines = {}

--ASS parser
function LoadASS(content)
	--Unicode width in UTF8
	local function charwidth(text, i)
		local byte = text:byte(i or 1)
		--[[
		U-00000000 EU-0000007F: 	0xxxxxxx
		U-00000080 EU-000007FF: 	110xxxxx 10xxxxxx
		U-00000800 EU-0000FFFF: 	1110xxxx 10xxxxxx 10xxxxxx
		U-00010000 EU-001FFFFF: 	11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
		U-00200000 EU-03FFFFFF: 	111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
		U-04000000 EU-7FFFFFFF: 	1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
		]]--
		if not byte then
			return 1
		elseif byte < 128 then
			return 1
		elseif byte < 224 then
			return 2
		elseif byte < 240 then
			return 3
		elseif byte < 248 then
			return 4
		elseif byte < 252 then
			return 5
		else
			return 6
		end
	end
	--Searching patterns
	local meta_comment = ";%s*(.*)"
	local meta_title = "Title:%s*(.*)"
	local meta_scripttype = "ScriptType:%s*(.*)"
	local meta_wrapstyle = "WrapStyle:%s*(%d+)"
	local meta_width = "PlayResX:%s*(%d+)"
	local meta_height = "PlayResY:%s*(%d+)"
	local meta_scalebands = "ScaledBorderAndShadow:%s*(.*)"
	local meta_ratio = "Video Aspect Ratio:%s*([%d%.]+)"
	local meta_zoom = "Video Zoom:%s*(%d+)"
	local meta_position = "Video Position:%s*(%d+)"
	local meta_laststyle = "Last Style Storage:%s*(.*)"
	local meta_audio = "Audio File:%s*(.*)"
	local meta_video = "Video File:%s*(.*)"
	local style_vals = "Style:%s*(.-),(.-),([%d%-]+),(.-),(.-),(.-),(.-),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-%.]+),([%d%-%.]+),([%d%-]+),([%d%-%.]+),([%d%-%.]+),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-]+),([%d%-]+)"
	local dialog_line = "Dialogue:%s*([%d%-]+),(.-),(.-),(.-),(.-),([%d%-]+),([%d%-]+),([%d%-]+),(.-),(.*)"
	local comment_line = "Comment:%s*([%d%-]+),(.-),(.-),(.-),(.-),([%d%-]+),([%d%-]+),([%d%-]+),(.-),(.*)"
	--Parsing line
	local function LoadLine(line)
		--META
		if line:find(meta_comment) then
			if meta.comment then
				meta.comment = meta.comment .. "\n" .. line:match(meta_comment)
			else
				meta.comment = line:match(meta_comment)
			end
		end
		if line:find(meta_title) then
			meta.title = line:match(meta_title)
		end
		if line:find(meta_scripttype) then
			meta.script_type = line:match(meta_scripttype)
		end
		if line:find(meta_wrapstyle) then
			meta.wrap_style = tonumber(line:match(meta_wrapstyle))
		end
		if line:find(meta_width) then
			meta.width = tonumber(line:match(meta_width))
		end
		if line:find(meta_height) then
			meta.height = tonumber(line:match(meta_height))
		end
		if line:find(meta_scalebands) then
			meta.scaled_border_and_shadow = line:match(meta_scalebands)
			meta.scaled_border_and_shadow = meta.scaled_border_and_shadow == "yes"
		end
		if line:find(meta_ratio) then
			meta.ratio = tonumber(line:match(meta_ratio))
		end
		if line:find(meta_zoom) then
			meta.zoom = tonumber(line:match(meta_zoom))
		end
		if line:find(meta_position) then
			meta.position = tonumber(line:match(meta_position))
		end
		if line:find(meta_laststyle) then
			meta.last_style = line:match(meta_laststyle)
		end
		if line:find(meta_audio) then
			meta.audio = line:match(meta_audio)
		end
		if line:find(meta_video) then
			meta.video = line:match(meta_video)
		end
		--STYLES
		if line:find(style_vals) then
			local style = {}
			style.name, style.fontname, style.fontsize,
			style.color1, style.color2, style.color3, style.color4,
			style.bold, style.italic, style.underline, style.strikeout,
			style.scale_x, style.scale_y, style.spacing, style.angle,
			style.borderstyle, style.outline, style.shadow,
			style.alignment, style.margin_l, style.margin_r, style.margin_v,
			style.encoding =
			line:match(style_vals)
			if style.encoding then
				style.fontsize = tonumber(style.fontsize)
				local a, b, g, r = style.color1:match("&H(%x%x)(%x%x)(%x%x)(%x%x)")
				if r then
					style.color1 = string.format("&H%02x%02x%02x&", tonumber(b,16), tonumber(g,16), tonumber(r,16))
					style.alpha1 = string.format("&H%02x&", tonumber(a,16))
				else
					style.color1 = nil
				end
				a, b, g, r = style.color2:match("&H(%x%x)(%x%x)(%x%x)(%x%x)")
				if r then
					style.color2 = string.format("&H%02x%02x%02x&", tonumber(b,16), tonumber(g,16), tonumber(r,16))
					style.alpha2 = string.format("&H%02x&", tonumber(a,16))
				else
					style.color2 = nil
				end
				a, b, g, r = style.color3:match("&H(%x%x)(%x%x)(%x%x)(%x%x)")
				if r then
					style.color3 = string.format("&H%02x%02x%02x&", tonumber(b,16), tonumber(g,16), tonumber(r,16))
					style.alpha3 = string.format("&H%02x&", tonumber(a,16))
				else
					style.color3 = nil
				end
				a, b, g, r = style.color4:match("&H(%x%x)(%x%x)(%x%x)(%x%x)")
				if r then
					style.color4 = string.format("&H%02x%02x%02x&", tonumber(b,16), tonumber(g,16), tonumber(r,16))
					style.alpha4 = string.format("&H%02x&", tonumber(a,16))
				else
					style.color4 = nil
				end
				style.bold = tonumber(style.bold) and tonumber(style.bold) == -1 or false
				style.italic = tonumber(style.italic) and tonumber(style.italic) == -1 or false
				style.underline = tonumber(style.underline) and tonumber(style.underline) == -1 or false
				style.strikeout = tonumber(style.strikeout) and tonumber(style.strikeout) == -1 or false
				style.scale_x = tonumber(style.scale_x)
				style.scale_y = tonumber(style.scale_y)
				style.spacing = tonumber(style.spacing)
				style.angle = tonumber(style.angle)
				style.borderstyle = tonumber(style.borderstyle) and tonumber(style.borderstyle) == 3 or false
				style.outline = tonumber(style.outline)
				style.shadow = tonumber(style.shadow)
				style.alignment = tonumber(style.alignment)
				style.margin_l = tonumber(style.margin_l)
				style.margin_r = tonumber(style.margin_r)
				style.margin_v = tonumber(style.margin_v)
				style.encoding = tonumber(style.encoding)
				table.insert(styles, style)
			end
		end
		--LINES
		local c_line, d_line = line:find(comment_line), line:find(dialog_line)
		if c_line or d_line then
			local dialog = {}
			if d_line then
				dialog.comment = false
				dialog.layer, dialog.start_time, dialog.end_time, dialog.style, dialog.actor, dialog.margin_r, dialog.margin_l, dialog.margin_v, dialog.effect, dialog.k_text =
				line:match(dialog_line)
			else
				dialog.comment = true
				dialog.layer, dialog.start_time, dialog.end_time, dialog.style, dialog.actor, dialog.margin_r, dialog.margin_l, dialog.margin_v, dialog.effect, dialog.k_text =
				line:match(comment_line)
			end
			if dialog.k_text then
				dialog.layer = tonumber(dialog.layer)
				local h, m, s, ms = dialog.start_time:match("(%d+):(%d+):(%d+).(%d+)")
				if ms then
					dialog.start_time = ms * 10 + s * 1000 + m * 1000 * 60 + h * 1000 * 60 *60
				else
					dialog.start_time = nil
				end
				h, m, s, ms = dialog.end_time:match("(%d+):(%d+):(%d+).(%d+)")
				if ms then
					dialog.end_time = ms * 10 + s * 1000 + m * 1000 * 60 + h * 1000 * 60 *60
				else
					dialog.end_time = nil
				end
				dialog.margin_r = tonumber(dialog.margin_r)
				dialog.margin_l = tonumber(dialog.margin_l)
				dialog.margin_v = tonumber(dialog.margin_v)
				table.insert(lines, dialog)
			end
		end
	end
	--Extend dialog data
	local function ExtendASS()
		--Process
		for li, line in ipairs(lines) do
			--Line basics
			if line.start_time and line.end_time then
				line.duration = line.end_time - line.start_time
			end
			line.i = li
			for si, style in ipairs(styles) do
				if line.style == style.name then
					line.styleref = style
				end
			end
			line.text = line.k_text:gsub("{(.-)}", "")
			--Sylables basics
			line.syls = {}
			local last_time, word_i = 0, 1
			for pre_tag, t, post_tag, text in line.k_text:gmatch("{(.-)\\[kK][of]?(%d+)(.-)}([^{]*)") do
				local syl = {}
				syl.duration = t * 10
				syl.start_time = last_time
				syl.end_time = syl.start_time + syl.duration
				last_time = syl.end_time
				syl.i = #line.syls + 1
				if text:sub(1,1) == " " then word_i = word_i + 1 end
				syl.word_i = word_i
				if text:sub(text:len(), text:len()) == " " then word_i = word_i + 1 end
				syl.text = text:gsub("^%s*(.-)%s*$", "%1")
				syl.prespace = text:match("^(%s*)"):len()
				if syl.text == "" then
					syl.postspace = 0
				else
					syl.postspace = text:match("(%s*)$"):len()
				end
				local fx = string.match(pre_tag .. post_tag, "\\%-([^\\]*)")
				if fx then
					syl.inline_fx = fx
				else
					syl.inline_fx = ""
				end
				table.insert(line.syls, syl)
			end
			--Characters basics
			line.chars = {}
			local ci, word_i = 1, 1
			while ci <= line.text:len() do
				local char = {}
				local cw = charwidth(line.text, ci)
				char.i = #line.chars + 1
				char.start_time = 0
				char.end_time = line.duration
				char.duration = line.duration
				char.text = line.text:sub(ci, ci + cw - 1)
				if char.text == " " then char.word_i = 0 else char.word_i = word_i end
				if char.text == " " and char.i ~= 1 and line.chars[char.i-1] ~= " " then word_i = word_i + 1 end
				char.syl_i = 1
				table.insert(line.chars, char)
				ci = ci + cw
			end
			local char_i = 1
			for si, syl in ipairs(line.syls) do
				--Prespace
				for p = 0, syl.prespace - 1 do
					line.chars[char_i + p].start_time = syl.start_time
					line.chars[char_i + p].end_time = syl.end_time
					line.chars[char_i + p].duration = syl.duration
					line.chars[char_i + p].syl_i = si
				end
				char_i = char_i + syl.prespace
				--Characters
				ci = 1
				while ci <= syl.text:len() do
					line.chars[char_i].start_time = syl.start_time
					line.chars[char_i].end_time = syl.end_time
					line.chars[char_i].duration = syl.duration
					line.chars[char_i].syl_i = si
					ci = ci + charwidth(syl.text, ci)
					char_i = char_i + 1
				end
				--Postspace
				for p = 0, syl.postspace - 1 do
					line.chars[char_i + p].start_time = syl.start_time
					line.chars[char_i + p].end_time = syl.end_time
					line.chars[char_i + p].duration = syl.duration
					line.chars[char_i + p].syl_i = si
				end
				char_i = char_i + syl.postspace
			end
			--Words basics
			line.words = {}
			for text in line.text:gmatch("%s*[^%s]+%s*") do
				local word = {}
				word.i = #line.words + 1
				word.start_time = 0
				word.end_time = line.duration
				word.duration = line.duration
				word.text = text:gsub("^%s*(.-)%s*$", "%1")
				word.prespace = text:match("^(%s*)"):len()
				word.postspace = text:match("(%s*)$"):len()
				table.insert(line.words, word)
			end
			--Sizes
			if line.styleref and meta.width and meta.height then
				local space_width = TextExtents(" ", line.styleref)
				--Line size
				line.width, line.height, line.ascent, line.descent, line.intlead, line.extlead = TextExtents(line.text, line.styleref)
				--Line position
				--y
				if line.styleref.alignment > 6 then
					line.top = math.max(line.margin_v, line.styleref.margin_v)
					line.middle = line.top + line.height / 2
					line.bottom = line.top + line.height
					line.y = line.top
				elseif line.styleref.alignment > 3 then
					line.top = meta.height / 2 - line.height / 2
					line.middle = line.top + line.height / 2
					line.bottom = line.top + line.height
					line.y = line.middle
				else
					line.top = meta.height - math.max(line.margin_v, line.styleref.margin_v) - line.height
					line.middle = line.top + line.height / 2
					line.bottom = line.top + line.height
					line.y = line.bottom
				end
				--x
				if math.mod(line.styleref.alignment-1, 3) == 0 then
					line.left = math.max(line.margin_l, line.styleref.margin_l)
					line.center = line.left + line.width / 2
					line.right = line.left + line.width
					line.x = line.left
				elseif math.mod(line.styleref.alignment-2, 3) == 0 then
					line.left = meta.width / 2 - line.width / 2
					line.center = line.left + line.width / 2
					line.right = line.left + line.width
					line.x = line.center
				else
					line.left = meta.width - math.max(line.margin_r, line.styleref.margin_r) - line.width
					line.center = line.left + line.width / 2
					line.right = line.left + line.width
					line.x = line.right
				end
				--Sylables sizes
				local max_width = 0
				local sum_height = 0
				for si, syl in ipairs(line.syls) do
					syl.width, syl.height, syl.ascent, syl.descent, syl.intlead, syl.extlead = TextExtents(syl.text, line.styleref)
					max_width = math.max(max_width, syl.width)
					sum_height = sum_height + syl.height
				end
				--Sylables positions
				local cur_x = line.left
				local cur_y = meta.height / 2 - sum_height / 2
				--y
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for si, syl in ipairs(line.syls) do
						syl.top = line.top
						syl.middle = line.middle
						syl.bottom = line.bottom
						syl.y = line.y
					end
				else
					for si, syl in ipairs(line.syls) do
						syl.top = cur_y
						syl.middle = syl.top + syl.height / 2
						syl.bottom = syl.top + syl.height
						syl.y = syl.middle
						cur_y = cur_y + syl.height
					end
				end
				--x
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for si, syl in ipairs(line.syls) do
						cur_x = cur_x + syl.prespace * space_width
						syl.left = cur_x
						syl.center = syl.left + syl.width / 2
						syl.right = syl.left + syl.width
						if math.mod(line.styleref.alignment-1, 3) == 0 then
							syl.x = syl.left
						elseif math.mod(line.styleref.alignment-2, 3) == 0 then
							syl.x = syl.center
						else
							syl.x = syl.right
						end
						cur_x = cur_x + syl.width + syl.postspace * space_width
					end
				else
					local x_fix = 0
					for si, syl in ipairs(line.syls) do
						x_fix = (max_width - syl.width) / 2
						if line.styleref.alignment == 4 then
							syl.left = line.left + x_fix
							syl.center = syl.left + syl.width / 2
							syl.right = syl.left + syl.width
							syl.x = syl.left
						elseif line.styleref.alignment == 5 then
							syl.left = meta.width / 2 - syl.width / 2
							syl.center = syl.left + syl.width / 2
							syl.right = syl.left + syl.width
							syl.x = syl.center
						else
							syl.left = line.right - syl.width - x_fix
							syl.center = syl.left + syl.width / 2
							syl.right = syl.left + syl.width
							syl.x = syl.right
						end
					end
				end
				--Characters sizes
				max_width = 0
				sum_height = 0
				for ci, char in ipairs(line.chars) do
					char.width, char.height, char.ascent, char.descent, char.intlead, char.extlead = TextExtents(char.text, line.styleref)
					max_width = math.max(max_width, char.width)
					sum_height = sum_height + char.height
				end
				--Characters positions
				cur_x = line.left
				cur_y = meta.height / 2 - sum_height / 2
				--y
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for ci, char in ipairs(line.chars) do
						char.top = line.top
						char.middle = line.middle
						char.bottom = line.bottom
						char.y = line.y
					end
				else
					for ci, char in ipairs(line.chars) do
						char.top = cur_y
						char.middle = char.top + char.height / 2
						char.bottom = char.top + char.height
						char.y = char.middle
						cur_y = cur_y + char.height
					end
				end
				--x
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for ci, char in ipairs(line.chars) do
						char.left = cur_x
						char.center = char.left + char.width / 2
						char.right = char.left + char.width
						if math.mod(line.styleref.alignment-1, 3) == 0 then
							char.x = char.left
						elseif math.mod(line.styleref.alignment-2, 3) == 0 then
							char.x = char.center
						else
							char.x = char.right
						end
						cur_x = cur_x + char.width
					end
				else
					local x_fix = 0
					for ci, char in ipairs(line.chars) do
						x_fix = (max_width - char.width) / 2
						if line.styleref.alignment == 4 then
							char.left = line.left + x_fix
							char.center = char.left + char.width / 2
							char.right = char.left + char.width
							char.x = char.left
						elseif line.styleref.alignment == 5 then
							char.left = meta.width / 2 - char.width / 2
							char.center = char.left + char.width / 2
							char.right = char.left + char.width
							char.x = char.center
						else
							char.left = line.right - char.width - x_fix
							char.center = char.left + char.width / 2
							char.right = char.left + char.width
							char.x = char.right
						end
					end
				end
				--Words sizes
				max_width = 0
				sum_height = 0
				for wi, word in ipairs(line.words) do
					word.width, word.height, word.ascent, word.descent, word.intlead, word.extlead = TextExtents(word.text, line.styleref)
					max_width = math.max(max_width, word.width)
					sum_height = sum_height + word.height
				end
				--Words positions
				cur_x = line.left
				cur_y = meta.height / 2 - sum_height / 2
				--y
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for si, word in ipairs(line.words) do
						word.top = line.top
						word.middle = line.middle
						word.bottom = line.bottom
						word.y = line.y
					end
				else
					for si, word in ipairs(line.words) do
						word.top = cur_y
						word.middle = word.top + word.height / 2
						word.bottom = word.top + word.height
						word.y = word.middle
						cur_y = cur_y + word.height
					end
				end
				--x
				if line.styleref.alignment > 6 or line.styleref.alignment < 4 then
					for si, word in ipairs(line.words) do
						cur_x = cur_x + word.prespace * space_width
						word.left = cur_x
						word.center = word.left + word.width / 2
						word.right = word.left + word.width
						if math.mod(line.styleref.alignment-1, 3) == 0 then
							word.x = word.left
						elseif math.mod(line.styleref.alignment-2, 3) == 0 then
							word.x = word.center
						else
							word.x = word.right
						end
						cur_x = cur_x + word.width + word.postspace * space_width
					end
				else
					local x_fix = 0
					for si, word in ipairs(line.words) do
						x_fix = (max_width - word.width) / 2
						if line.styleref.alignment == 4 then
							word.left = line.left + x_fix
							word.center = word.left + word.width / 2
							word.right = word.left + word.width
							word.x = word.left
						elseif line.styleref.alignment == 5 then
							word.left = meta.width / 2 - word.width / 2
							word.center = word.left + word.width / 2
							word.right = word.left + word.width
							word.x = word.center
						else
							word.left = line.right - word.width - x_fix
							word.center = word.left + word.width / 2
							word.right = word.left + word.width
							word.x = word.right
						end
					end
				end
			end
		end
	end
	--Adds in- & outfade time to dialog lines
	local function InOutTimes(offset)
		if type(offset) ~= "number" then
			error("number expected", 2)
		end

		--Collect style names and fitting lines
		local line_styles = {}
		for li, line in ipairs(lines) do
			if not line_styles[line.style] then
				line_styles[line.style] = {}
			end
			table.insert(line_styles[line.style], {start_time = line.start_time, end_time = line.end_time, i = line.i})
		end
		--Sort by time
		for si, style in pairs(line_styles) do
			--Collect and order start times
			local start_times = {}
			for li, line in ipairs(style) do
				table.insert(start_times, line.start_time)
			end
			table.sort(start_times)
			--Order lines by start times
			local style_cpy = table.copy(style)
			for ti, start_time in ipairs(start_times) do
				for li, line in ipairs(style_cpy) do
					if line ~= nil and start_time == line.start_time then
						style[ti] = table.copy(line)
						line = nil
						break
					end
				end
			end
		end
		--Insert calculated in- & outfade times
		for si, style in pairs(line_styles) do
			for li, line in ipairs(style) do
				lines[line.i].infade = (li == 1) and offset or line.start_time - style[li-1].end_time
				if lines[line.i].infade < 0 then lines[line.i].infade = 0 end
				lines[line.i].outfade = (li == #style) and offset or style[li+1].start_time - line.end_time
				if lines[line.i].outfade < 0 then lines[line.i].outfade = 0 end
			end
		end
	end
	--Check + read file
	if type(content) == "string" then
		for line in content:gmatch("[^\n]+") do
			LoadLine(line)
		end
		ExtendASS()
		InOutTimes(1000.1)
	else
		error("string expected", 2)
	end
end
