--Prints value to GUI or file
function print(value, target)
	if (type(target) ~= "string") and (type(target) ~= "nil") then
		error("any value and optional string expected", 2)
	end

	if not target then
		gui.print(tostring(value))
	else
		local f, err = io.open(target, "a")
		if err then
			error(err, 2)
		end
		f:write(tostring(value))
		f:close()
	end
end

--Sets progress bar
function ProgressStatus(pct)
	if type(pct) ~= "number" then
		error("number expected", 2)
	end

	local pct = (pct < 0 and 0) or (pct > 1 and 1) or pct
	gui.progress(pct * 100)
end

--Converts an ASS dialog line into text and writes it to temporary output file
function Output(line)
	--Test line
	if type(line) ~= "table" then
		error("table expected", 2)
	end
	local check = {
		comment = false,
		layer = false,
		start_time = false,
		end_time = false,
		style = false,
		name = false,
		marginl = false,
		marginr = false,
		marginv = false,
		effect = false,
		text = false
	}
	for i, v in pairs(line) do
		if tostring(i) == "comment" and type(v) == "boolean" then
			check.comment = true
		elseif tostring(i) == "layer" and type(v) == "number" then
			check.layer = true
		elseif tostring(i) == "start_time" and type(v) == "number" then
			check.start_time = true
		elseif tostring(i) == "end_time" and type(v) == "number" then
			check.end_time = true
		elseif tostring(i) == "style" and type(v) == "string" then
			check.style = true
		elseif tostring(i) == "actor" and type(v) == "string" then
			check.name = true
		elseif tostring(i) == "margin_l" and type(v) == "number" then
			check.marginl = true
		elseif tostring(i) == "margin_r" and type(v) == "number" then
			check.marginr = true
		elseif tostring(i) == "margin_v" and type(v) == "number" then
			check.marginv = true
		elseif tostring(i) == "effect" and type(v) == "string" then
			check.effect = true
		elseif tostring(i) == "text" and type(v) == "string" then
			check.text = true
		end
	end
	for i, v in pairs(check) do
		if not v then
			error("valid table expected", 2)
		end
	end

	--Convert line
	local function timestamp(t)
		local temp_time = t
		--Divisors
		local hour_divisor = 60 * 60 * 1000
		local minute_divisor = 60 * 1000
		local second_divisor = 1000
		--Hours
		local h = math.floor( temp_time / hour_divisor )
		temp_time = math.mod( temp_time, hour_divisor )
		--Minutes
		local m = math.floor( temp_time / minute_divisor )
		temp_time = math.mod( temp_time, minute_divisor)
		--Seconds
		local s = math.floor( temp_time / second_divisor )
		temp_time = math.mod( temp_time, second_divisor)
		--Milliseconds
		local ms = math.floor(temp_time / 10)
		--Timestamp
		return string.format("%d:%02d:%02d.%02d", h, m, s, ms)
	end
	local text = string.format("\n%s: %d,%s,%s,%s,%s,%04d,%04d,%04d,%s,%s",
								line.comment and "Comment" or "Dialogue",
								line.layer,
								timestamp(line.start_time),
								timestamp(line.end_time),
								line.style,
								line.actor,
								line.margin_l,
								line.margin_r,
								line.margin_v,
								line.effect,
								line.text)
	--Write line
	gui.writefile(text, 1)
end
