--Rectangle
function Rectangle(w, h)
	if type(w) ~= "number" or type(h) ~= "number" then
		error("number and number expected", 2)
	end

	return string.format("m 0 0 l %d 0 %d %d 0 %d", w, w, h, h)
end

--Triangle
function Triangle(size)
	if type(size) ~= "number" then
		error("number expected", 2)
	end

	local h = math.sqrt(3) / 2 * size
	local my = ((h / 2) / (size * 0.75)) * (size / 2)
	local base = -(h - 2 * my) / 2
	print(base)
	return string.format("m %d %d l 0 %d %d %d", size/2, base, base+h, size, base+h )
end

--Ellipse
function Ellipse(w, h)
	if type(w) ~= "number" or type(h) ~= "number" then
		error("number and number expected", 2)
	end

	return string.format("m 0 %d b 0 %d 0 0 %d 0 %d 0 %d 0 %d %d %d %d %d %d %d %d %d %d 0 %d 0 %d", h/2, h/2, w/2, w/2, w, w, h/2, w, h/2, w, h, w/2, h, w/2, h, h, h/2)
end

--Star
function Star(edges, inner_size, outer_size)
	if type(edges) ~= "number" or type(inner_size) ~= "number" or type(outer_size) ~= "number" or
		edges < 2 then
		error("valid 3 numbers expected", 2)
	end

	--Util
	local function z_rot(x, y, a)
		local new_x = x * math.cos(a) - y * math.sin(a)
		local new_y = x * math.sin(a) + y * math.cos(a)
		return new_x, new_y
	end
	--Draw
	local shape = string.format("m 0 %d l", -outer_size)
	for i = 1, edges do
		--Inner edge
		local pct = (i-0.5) / edges
		local x, y = z_rot(0, -inner_size, math.pi*2*pct)
		shape = shape .. string.format(" %d %d", x, y)
		--Outer edge
		pct = i / edges
		x, y = z_rot(0, -outer_size, math.pi*2*pct)
		shape = shape .. string.format(" %d %d", x, y)
	end
	--Minimum x and y range
	local min_x, min_y = 0, 0
	local function search_min(x, y)
		min_x, min_y = math.min(min_x, tonumber(x)), math.min(min_y, tonumber(y))
	end
	shape:gsub("(%-?%d+) (%-?%d+)", search_min)
	--Shift to positive numbers
	local function shift(x, y)
		return string.format("%d %d", x-min_x, y-min_y)
	end
	shape = shape:gsub("(%-?%d+) (%-?%d+)", shift)
	return shape
end

--Glance
function Glance(edges, inner_size, outer_size)
	if type(edges) ~= "number" or type(inner_size) ~= "number" or type(outer_size) ~= "number" or
		edges < 2 then
		error("valid 3 numbers expected", 2)
	end

	--Util
	local function z_rot(x, y, a)
		local new_x = x * math.cos(a) - y * math.sin(a)
		local new_y = x* math.sin(a) + y* math.cos(a)
		return new_x, new_y
	end
	--Draw
	local shape = string.format("m 0 %d b", -outer_size)
	for i = 1, edges do
		--Inner edge
		pct = (i-0.5) / edges
		x, y = z_rot(0, -inner_size, math.pi*2*pct)
		for ii = 1, 2 do
			shape = shape .. string.format(" %d %d", x, y)
		end
		--Outer edge
		pct = i / edges
		x, y = z_rot(0, -outer_size, math.pi*2*pct)
		shape = shape .. string.format(" %d %d", x, y)
	end
	--Minimum x and y range
	local min_x, min_y = 0, 0
	local function search_min(x, y)
		min_x, min_y = math.min(min_x, tonumber(x)), math.min(min_y, tonumber(y))
	end
	shape:gsub("(%-?%d+) (%-?%d+)", search_min)
	--Shift to positive numbers
	local function shift(x, y)
		return string.format("%d %d", x-min_x, y-min_y)
	end
	shape = shape:gsub("(%-?%d+) (%-?%d+)", shift)
	return shape
end

--Heart
function Heart(size, offset)
	if type(size) ~= "number" or type(offset) ~= "number" then
		error("number expected", 2)
	end

	--Size
	local function sizer(num)
		num = num / 30 * size
		return math.floor(num+0.5)
	end
	local shape = string.gsub("m 15 30 b 27 22 30 18 30 14 30 8 22 0 15 10 8 0 0 8 0 14 0 18 3 22 15 30", "%d+", sizer)
	--Center y offset
	local i = 0
	local function y_shift(y)
		i = i + 1
		if i == 14 then
			return math.floor(y + offset)
		else
			return y
		end
	end
	shape = shape:gsub("%d+", y_shift)
	return shape
end
