--[[
U-00000000 EU-0000007F: 	0xxxxxxx
U-00000080 EU-000007FF: 	110xxxxx 10xxxxxx
U-00000800 EU-0000FFFF: 	1110xxxx 10xxxxxx 10xxxxxx
U-00010000 EU-001FFFFF: 	11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
U-00200000 EU-03FFFFFF: 	111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
U-04000000 EU-7FFFFFFF: 	1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
]]--

--Unicode character length
function string.ucharrange(s, i)
	if type(s) ~= "string" or type(i) ~= "number" or i < 1 or i > s:len() then
		error("string and valid number expected", 2)
	end

	local byte = s:byte(1)
	if byte < 192 then
		return 1
	elseif byte < 224 then
		return 2
	elseif byte < 240 then
		return 3
	elseif byte < 248 then
		return 4
	elseif byte < 252 then
		return 5
	else
		return 6
	end
end

--Iterator function through unicode characters of a string
function string.uchars(s)
	if type(s) ~= "string" then
		error("string expected", 2)
	end

	local char_i, s_pos = 0, 1
	local function itor()
		if s_pos > s:len() then
			return nil
		end

		local char_w = string.ucharrange(s, s_pos)
		local cur_pos = s_pos
		s_pos = s_pos + char_w
		char_i = char_i + 1
		return char_i, s:sub(cur_pos, cur_pos + char_w - 1)
	end
	return itor
end

--Unicode string length
function string.ulen(s)
	if type(s) ~= "string" then
		error("string expected", 2)
	end

	local n = 0
	for _ in string.uchars(s) do
		n = n + 1
	end
	return n
end

--Converts an unicode hex code to an UTF8 string
function string.unicode_to_utf8(unicode)
	if type(unicode) ~= "string" then
		error("string expected", 2)
	end

	--Utils
	local unicode_tables = {
		"0xxxxxxx",
		"110xxxxx10xxxxxx",
		"1110xxxx10xxxxxx10xxxxxx",
		"11110xxx10xxxxxx10xxxxxx10xxxxxx",
		"111110xx10xxxxxx10xxxxxx10xxxxxx10xxxxxx",
		"1111110x10xxxxxx10xxxxxx10xxxxxx10xxxxxx10xxxxxx"
	}
	local function to_nibble(hex_code)
		local hex_num = tonumber(hex_code, 16)
		local nibble = ""
		for i = 1, 4 do
			local remainder = math.mod(hex_num, 2)
			hex_num = math.floor(hex_num / 2)
			nibble = remainder .. nibble
		end
		return nibble
	end

	--Erase 0s at the beginning
	local unicode = string.gsub(unicode, "^0*", "")

	--Unicode to bits
	local bitcode = ""
	for si = unicode:len(), 1, -1 do
		bitcode = to_nibble(unicode:sub(si, si)) .. bitcode
	end

	--Choose unicode table for insertion
	local unicode_table
	local uni_len = unicode:len()
	if uni_len < 3 and (uni_len ~= 2 or tonumber(unicode:sub(1,1), 16) < 8) then
		unicode_table = unicode_tables[1]
	elseif uni_len < 4 and (uni_len ~= 3 or tonumber(unicode:sub(1,1), 16) < 8) then
		unicode_table = unicode_tables[2]
	elseif uni_len < 5 then
		unicode_table = unicode_tables[3]
	elseif uni_len < 7 and (uni_len ~= 6 or tonumber(unicode:sub(1,1), 16) < 2) then
		unicode_table = unicode_tables[4]
	elseif uni_len < 8 and (uni_len ~= 7 or tonumber(unicode:sub(1,1), 16) < 4) then
		unicode_table = unicode_tables[5]
	else
		unicode_table = unicode_tables[6]
	end

	--Insert unicode bits to unicode table
	local bitcode_result = ""
	for i = unicode_table:len(), 1, -1 do
		local table_char = unicode_table:sub(i,i)
		if table_char == "x" then
			local uni_len = bitcode:len()
			if uni_len < 1 then
				bitcode_result = "0" .. bitcode_result
			else
				local uni_char = bitcode:sub( uni_len, uni_len )
				bitcode_result = uni_char .. bitcode_result
				bitcode = bitcode:sub(1, uni_len - 1)
			end
		else
			bitcode_result = table_char .. bitcode_result
		end
	end

	--Final bitcode to utf8 characters
	local utf8 = ""
	for i = 1, math.floor(bitcode_result:len() / 8) do
		local len = bitcode_result:len()
		local byte = bitcode_result:sub(len-7, len)
		utf8 = string.char(tonumber(byte, 2)) .. utf8
		bitcode_result = bitcode_result:sub(1, len-8)
	end
	return utf8
end
