--Deep table copy
function table.copy(old_t)
	if type(old_t) ~= "table" then
		error("table expected", 2)
	end

	local new_t = {}
	for index, value in pairs(old_t) do
		if type(value) == "table" then
		    new_t[index] = table.copy(value)
		else
            new_t[index] = value
		end
	end
	return new_t
end

--Table size (numeric + literal)
function table.max(t)
	if type(t) ~= "table" then
		error("table expected", 2)
	end

	local n, index = 0, nil
	index = next(t, index)
	while index do
		n = n + 1
		index = next(t, index)
	end
	return n
end

--Fills a table with one value
function table.fill(t, count, value)
	if type(t) ~= "table" or type(count) ~= "number" then
		error("table, number and any value expected", 2)
	end

	for i = 1, count do
		t[i] = value
	end
end

--Appends table two on table one
function table.append(t1, t2)
	if type(t1) ~= "table" or type(t2) ~= "table" then
		error("table and table expected", 2)
	end

	for i, v in pairs(t2) do
		if tonumber(i) then
			table.insert(t1, v)
		else
			t1[i] = v
		end
	end
end

--Whole table content to a string
function table.tostring(t)
	if type(t) ~= "table" then
		error("table expected", 2)
	end

	local temp_text = ""
	local function table_print(t, space)
		for i, v in pairs(t) do
			if type(i) == "string" then
				i = "\"" .. i .. "\""
			end
			if type(v) == "string" then
				v = "\"" .. v .. "\""
			end
			temp_text = temp_text .. string.format("%s[%s] = %s\n", space, tostring(i), tostring(v))
			if type(v) == "table" then
				table_print(v, space .. "\t")
			end
		end
	end
	table_print(t, "\t")

	return temp_text
end
