#include "Lua.h"

Lua::Lua()
{
    L = luaL_newstate();
    luaL_openlibs(L);
}

Lua::~Lua()
{
    lua_close(L);
}

void Lua::register_func(char* name, int (*func) (lua_State*))
{
    lua_register(L, name, func);
}

void Lua::register_lib(const char *libname, const luaL_Reg *reg)
{
    luaL_register(L, libname, reg);
}

bool Lua::dofile(const wchar_t *wfilename)
{
    char *cfilename = W_to_C(wfilename);
    if(luaL_dofile(L, cfilename))
    {
        delete (cfilename);
        wchar_t *werror = C_to_W( lua_tostring(L, -1) );
        MessageBoxW(NULL, werror, L"Lua source", MB_OK | MB_ICONERROR);
        delete (werror);
        return false;
    }
    delete (cfilename);
    return true;
}

bool Lua::pcall(int args, int results)
{
    if( lua_pcall(L, args, results, 0) )
    {
        wchar_t *werror = C_to_W( lua_tostring(L, -1) );
        MessageBoxW(NULL, werror, L"Lua source", MB_OK | MB_ICONERROR);
        delete (werror);
        return false;
    }
    return true;
}

void Lua::getglobal(const char *name)
{
    lua_getglobal(L, name);
}

void Lua::pushstring(const char *s)
{
    lua_pushstring(L, s);
}

void Lua::pushnil()
{
    lua_pushnil(L);
}

bool Lua::isfunction(int i)
{
    return lua_isfunction(L, i);
}





