#include "generator.h"

Generator::Generator(FileManager *fm, GUI *p_wnd)
{
    first_time = clock();
    L = new Lua();
    this->fm = fm;

    pi.wnd = p_wnd;
    pi.output.clear();
    pi.line_count = 0;
}

Generator::~Generator()
{
    delete (L);
    pi.output.clear();
    _fcloseall();
}

bool Generator::run()
{
    //Lua environment
    luaL_reg reg[] =
    {
        {"progress", l_progress},
        {"print", l_print},
        {"writefile", l_writefile},
        {NULL, NULL}
    };
    L->register_lib("gui", reg);
    L->register_func("TextExtents", l_TextExtents);
    L->register_func("GetImage", l_GetImage);
    L->register_func("GetPathW", l_GetPathW);
    L->register_func("GetPixelsW", l_GetPixelsW);

    //Load .lua files in include folder
    _WDIR *dir; struct _wdirent *dir_file;
    wstring include_path = fm->path;
    include_path.append(L"include\\");
    dir = _wopendir(include_path.c_str());
    if (dir != NULL)
    {
        while ( (dir_file = _wreaddir(dir)) != NULL)
        {
            wstring file = include_path;
            file.append(dir_file->d_name);
            if ( file.length() > 4 && file.substr(file.length()-4) == L".lua")
                if( !L->dofile(file.c_str()) )
                {
                    _wclosedir(dir);
                    return false;
                }
        }
        _wclosedir(dir);
    }

    //ASS parsing
    L->getglobal("Init");
    if( L->isfunction(-1) )
    {
        if(fm->kt.empty())
            L->pushnil();
        else
        {
            char *cfile = W_to_C( fm->kt.c_str() );
            L->pushstring( cfile );
            delete (cfile);
        }
        if( !L->pcall(1, 0) )
            return false;
    }
    else
    {
        MessageBoxA(NULL, "Missing function \"Init\"!", "Lua source", MB_OK | MB_ICONERROR);
        return false;
    }

    //Execute effect file
    if( !L->dofile(fm->fx.c_str()) )
        return false;

    //Write final file
    FILE *final = _wfopen(fm->final.c_str(), L"w");
    if(!final)
    {
        MessageBoxA(NULL, "Couldn't create final file!", "Lua source", MB_OK | MB_ICONERROR);
        return false;
    }
    fputs(pi.output.c_str(), final);
    fclose(final);

    //Success
    return true;
}

unsigned long int Generator::get_lines()
{
    return pi.line_count;
}

long unsigned int Generator::get_time()
{
    clock_t last_time = clock();
    return (last_time - first_time) * 1000 / CLOCKS_PER_SEC;
}
