#include "luaenv.h"

//GETTER
void l_error(lua_State *L, const char *extramsg = 0)
{
	lua_Debug ar;
	lua_getstack(L, 0, &ar);
	lua_getinfo(L, "n", &ar);
	if(extramsg)
        luaL_error(L, "%s: %s", ar.name, extramsg);
    else
        luaL_error(L, "Bad arguments to '%s' function!", ar.name);
}

wchar_t* GetText(lua_State *L, int offset)
{
    return C_to_W(luaL_checkstring(L, offset));
}

ASSStyle GetStyle(lua_State *L, int offset)
{
    luaL_checktype(L, offset, LUA_TTABLE);

    ASSStyle style;
    ZeroMemory(&style, sizeof(style));

    lua_getfield(L, offset, "fontname");
    if (!lua_isstring(L, -1))
        l_error(L, "Invalid style table - FONTNAME!");
    wchar_t *wtext = C_to_W(lua_tostring(L, -1));
    if (wcslen(wtext) > 31)
    {
        delete (wtext);
        l_error(L, "Invalid style table - FONTNAME!");
    }
    wcscpy(style.fontname, wtext);
    delete (wtext);
    lua_pop(L, 1);

    lua_getfield(L, offset, "fontsize");
    if (!lua_isnumber(L, -1) || lua_tonumber(L, -1) < 0)
        l_error(L, "Invalid style table - FONTSIZE!");
    style.fontsize = lua_tonumber(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "bold");
    if (!lua_isboolean(L, -1))
        l_error(L, "Invalid style table - BOLD!");
    style.bold = lua_toboolean(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "italic");
    if (!lua_isboolean(L, -1))
        l_error(L, "Invalid style table - ITALIC!");
    style.italic = lua_toboolean(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "underline");
    if (!lua_isboolean(L, -1))
        l_error(L, "Invalid style table - UNDERLINE!");
    style.underline = lua_toboolean(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "strikeout");
    if (!lua_isboolean(L, -1))
        l_error(L, "Invalid style table - STRIKEOUT!");
    style.strikeout = lua_toboolean(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "encoding");
    if (!lua_isnumber(L, -1)  || lua_tonumber(L, -1) < 0)
        l_error(L, "Invalid style table - ENCODING!");
    style.encoding = lua_tonumber(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "scale_x");
    if (!lua_isnumber(L, -1) || lua_tonumber(L, -1) < 0)
        l_error(L, "Invalid style table - SCALE_X!");
    style.scale_x = lua_tonumber(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "scale_y");
    if (!lua_isnumber(L, -1) || lua_tonumber(L, -1) < 0)
        l_error(L, "Invalid style table - SCALE_Y!");
    style.scale_y = lua_tonumber(L, -1);
    lua_pop(L, 1);

    lua_getfield(L, offset, "spacing");
    if (!lua_isnumber(L, -1) || lua_tonumber(L, -1) < 0)
        l_error(L, "Invalid style table - SPACING!");
    style.spacing = lua_tonumber(L, -1);
    lua_pop(L, 1);

    return style;
}

HFONT SelectFont(HDC dc, ASSStyle style, unsigned short scale)
{
    LOGFONTW lf;
    ZeroMemory(&lf, sizeof(lf));
    lf.lfHeight = style.fontsize * scale;
    lf.lfWeight = style.bold ? FW_BOLD : FW_NORMAL;
    lf.lfItalic = style.italic;
    lf.lfUnderline = style.underline;
    lf.lfStrikeOut = style.strikeout;
    lf.lfCharSet = style.encoding;
    lf.lfOutPrecision = OUT_TT_PRECIS;
    lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    lf.lfQuality = ANTIALIASED_QUALITY;
    lf.lfPitchAndFamily = DEFAULT_PITCH|FF_DONTCARE;
    wcsncpy(lf.lfFaceName, style.fontname, 32);

    HFONT font = CreateFontIndirectW(&lf);
    SelectObject(dc, font);

    return font;
}

TXTExt TextExtents(wchar_t *wtext, ASSStyle style)
{
    //Text extents object
    TXTExt txtext;
    ZeroMemory(&txtext, sizeof(TXTExt));

    //Device context with font
    HDC dc = CreateCompatibleDC(0);
    SetMapMode(dc, MM_TEXT);
    HFONT font = SelectFont(dc, style, 64);

    //Text size
    SIZE sz;
    if (style.spacing)
    {
        for (unsigned long int i = 0; i < (unsigned long int)wcslen(wtext); i++)
        {
            GetTextExtentPoint32W(dc, &wtext[i], 1, &sz);
            txtext.width += sz.cx + style.spacing * 64;
            txtext.height = sz.cy;
        }
    }
    else
    {
        GetTextExtentPoint32W(dc, wtext, wcslen(wtext), &sz);
        txtext.width = sz.cx;
        txtext.height = sz.cy;
    }

    //Text metric
    TEXTMETRIC tm;
    GetTextMetrics(dc,&tm);
    txtext.ascent = tm.tmAscent;
    txtext.descent = tm.tmDescent;
    txtext.internal_lead = tm.tmInternalLeading;
    txtext.external_lead = tm.tmExternalLeading;

    //Size fix
    txtext.width = txtext.width * (style.scale_x / 100) / 64;
    txtext.height = txtext.height * (style.scale_y / 100) / 64;
    txtext.ascent = txtext.ascent * (style.scale_y / 100) / 64;
    txtext.descent = txtext.descent * (style.scale_y / 100) / 64;
    txtext.internal_lead = txtext.internal_lead * (style.scale_y / 100) / 64;
    txtext.external_lead = txtext.external_lead * (style.scale_y / 100) / 64;

    //Free device
    DeleteObject(dc);
    DeleteObject(font);

    //Return
    return txtext;
}

//SETTER
POINT DrawPath(HDC dc, wchar_t *shape, unsigned int scale)
{
    //Chunk number of shape command
    unsigned int pts_n = 1;
    for (unsigned int i = 0; i < wcslen(shape); i++)
        if (shape[i] == L' ')
            pts_n++;
    pts_n /= 2;
    //Shape points and types memory
    POINT *pts = new POINT[pts_n];
    BYTE *types = new BYTE[pts_n];

    //Path size (return value)
    POINT metric = {0, 0};

    //Read points and types
    BYTE cur_typ = PT_MOVETO; int p;
    unsigned int i = 0;
    for( wchar_t *token = wcstok(shape, L" "); token; token = wcstok(NULL, L" ") )
    {
        //Type?
        if ( wcscmp(token, L"m") == 0 )
            cur_typ = PT_MOVETO;
        else if ( wcscmp(token, L"l") == 0 )
            cur_typ = PT_LINETO;
        else if ( wcscmp(token, L"b") == 0 )
            cur_typ = PT_BEZIERTO;
        else
        {
            //x?
            p = _wtoi(token);
            if ( !(p == 0 && wcscmp(token, L"0") != 0) )
            {
                pts[i].x = p * scale;

                //2nd point?
                if ( (token = wcstok(NULL, L" ")) )
                {
                    //y?
                    p = _wtoi(token);
                    if ( !(p == 0 && wcscmp(token, L"0") != 0) )
                    {
                        pts[i].y = p * scale;

                        //Update width and height
                        if(pts[i].x > metric.x)
                            metric.x = pts[i].x;
                        if(pts[i].y > metric.y)
                            metric.y = pts[i].y;

                        //Append point
                        types[i] = cur_typ;
                        i++;
                    }
                }
            }
        }
    }
    //Last type closes path
    if (i>0) types[i-1] |= PT_CLOSEFIGURE;

    //Draw path
    BeginPath(dc);
    PolyDraw(dc, pts, types, i);
    EndPath(dc);

    //Return path size
    return metric;
}

void PathToLua(lua_State *L, HDC dc, unsigned short offset)
{
    //Curves to lines
    FlattenPath(dc);

    //Get path data
    unsigned int pts_n = GetPath(dc, NULL, NULL, 0);
    POINT *pts = new POINT[pts_n];
    BYTE *types = new BYTE[pts_n];
    GetPath(dc, pts, types, pts_n);

    //Set start point
    pts += offset;
    types += offset;

    //Return
    lua_newtable(L);
    for (unsigned int pts_i = 1; pts_i <= (pts_n-offset); pts_i++)
    {
        //Point
        lua_newtable(L);
        //Type
        if ( types[0] == PT_MOVETO )
            lua_pushstring(L, "m");
        else
            lua_pushstring(L, "l");
        lua_setfield(L, -2, "type");
        //x
        lua_pushnumber(L, pts->x);
        lua_rawseti(L, -2, 1);
        //y
        lua_pushnumber(L, pts->y);
        lua_rawseti(L, -2, 2);
        //z
        lua_pushnumber(L, 0);
        lua_rawseti(L, -2, 3);
        //Set point
        lua_rawseti(L, -2, pts_i);
        //Next point
        pts++; types++;
    }
}

void PixelsToLua(lua_State *L, HDC dc, POINT metric, unsigned int range)
{
    //Pixels table
    lua_newtable(L);
    //Through pixels
    unsigned long i = 0, c = 0, divisor = range*range;
    for (long y = 0; y < metric.y; y+= range)
    {
        for (long x = 0; x < metric.x; x+= range)
        {
            //Test
            c = 0;
            for(unsigned int yy = 0; yy < range; yy++)
                for(unsigned int xx = 0; xx < range; xx++)
                    c += GetRValue(GetPixel(dc, x+xx, y+yy));
            if(c>0)
            {
                //Pixel table
                lua_newtable(L);
                //Position
                lua_pushnumber(L, (double)x / range);
                lua_setfield(L, -2, "x");
                lua_pushnumber(L, (double)y / range);
                lua_setfield(L, -2, "y");
                //Color
                lua_pushnumber(L, (double)c / divisor);
                lua_setfield(L, -2, "a");
                //Insert pixel
                lua_rawseti(L, -2, ++i);
            }
        }
    }
}

//LUA FUNCTIONS
Proc_Info pi = {NULL, "", 0};

IMPLENT(progress)
{
    if(lua_gettop(L) == 1)
        pi.wnd->set_progressbar(luaL_checknumber(L, 1));
    else
        l_error(L);
    return 0;
}

IMPLENT(print)
{
    if(lua_gettop(L) == 1)
    {
        wchar_t *wmessage = GetText(L, 1);
        pi.wnd->log_append(wmessage);
        delete (wmessage);
    }else
        l_error(L);
    return 0;
}

IMPLENT(writefile)
{
    if(lua_gettop(L) == 1 || lua_gettop(L) == 2)
    {
        pi.output.append(luaL_checkstring(L, 1));
        if(lua_gettop(L) == 2)
            pi.line_count += luaL_checknumber(L, 2);
    }else
        l_error(L);
    return 0;
}

IMPLENT(TextExtents)
{
    if(lua_gettop(L) == 2)
    {
        ASSStyle style = GetStyle(L, 2);
        wchar_t *wtext = GetText(L, 1);

        TXTExt txtext = TextExtents(wtext, style);
        delete (wtext);

        lua_pushnumber(L, txtext.width);
        lua_pushnumber(L, txtext.height);
        lua_pushnumber(L, txtext.ascent);
        lua_pushnumber(L, txtext.descent);
        lua_pushnumber(L, txtext.internal_lead);
        lua_pushnumber(L, txtext.external_lead);
        return 6;
    }else
        l_error(L);
    return 0;
}

IMPLENT(GetImage)
{
    if(lua_gettop(L) == 1)
    {
        PNGImage *img = LoadPNG(GetText(L, 1));
        if(!img)
            return 0;

        //Image table
        lua_newtable(L);
        //Width & height
        lua_pushnumber(L, img->width);
        lua_setfield(L, -2, "width");
        lua_pushnumber(L, img->height);
        lua_setfield(L, -2, "height");
        //Data
        unsigned long int i = 0;
        for(unsigned int y = 0; y < img->height; y++)
            for(unsigned int x = 0; x < img->width; x++)
            {
                //Pixel table
                lua_newtable(L);
                //Coordinate
                lua_pushnumber(L, x);
                lua_setfield(L, -2, "x");
                lua_pushnumber(L, y);
                lua_setfield(L, -2, "y");
                //Color
                lua_pushnumber(L,img->data.at(i*4));
                lua_setfield(L, -2, "r");
                lua_pushnumber(L,img->data.at(i*4+1));
                lua_setfield(L, -2, "g");
                lua_pushnumber(L,img->data.at(i*4+2));
                lua_setfield(L, -2, "b");
                lua_pushnumber(L,img->data.at(i*4+3));
                lua_setfield(L, -2, "a");
                //Data index
                lua_rawseti(L, -2, ++i);
            }

        delete (img);
        return 1;
    }else
        l_error(L);
    return 0;
}

IMPLENT(GetPathW)
{
    //Text
    if(lua_gettop(L) == 2)
    {
        //Get text and style
        ASSStyle style = GetStyle(L, 2);
        wchar_t *wtext = GetText(L, 1);

        //Create context
        HDC dc = CreateCompatibleDC(0);
        SetMapMode(dc, MM_TEXT);

        //Set font
        HFONT font = SelectFont(dc, style, SCALE_FACTOR);

        //Create return table
        lua_newtable(L);
        unsigned int table_i = 0;

        //Iterate through characters
        wchar_t wc[2]; wc[1] = L'\0';
        double x = 0;
        for(unsigned long int i = 0; i < wcslen(wtext); i++)
        {
            //Draw character
            wc[0] = wtext[i];
            BeginPath(dc);
            TextOutW(dc, x, 0, wc, 1);
            EndPath(dc);
            x += TextExtents(wc, style).width * SCALE_FACTOR;
            //Insert character path in return table
            PathToLua(L, dc, 4);
            unsigned long int len = lua_objlen(L,-1);
            for(unsigned long int ii = 1; ii <= len; ii++)
            {
                lua_rawgeti(L, -1, ii);
                lua_rawseti(L, -3, ++table_i);
            }
            //Remove temporary character path table
            lua_pop(L, 1);
            //Clear context for next path
            AbortPath(dc);
        }

        //Free memory
        DeleteObject(dc);
        DeleteObject(font);
        delete (wtext);

        return 1;
    }
    //Shape
    else if(lua_gettop(L) == 1)
    {
        //Get text
        wchar_t *wtext = GetText(L, 1);

        //Create context
        HDC dc = CreateCompatibleDC(0);
        SetMapMode(dc, MM_TEXT);

        //Create path
        DrawPath(dc, wtext, SCALE_FACTOR);

        //Return path
        PathToLua(L, dc, 0);

        //Free memory
        DeleteObject(dc);
        delete (wtext);

        return 1;
    }else
        l_error(L);
    return 0;
}

IMPLENT(GetPixelsW)
{
    if(lua_gettop(L) == 1)
    {
        //Get text
        wchar_t *wtext = GetText(L, 1);

        //Create context
        HDC dc = CreateCompatibleDC(0);
        SetMapMode(dc, MM_TEXT);

        //Create path
        POINT metric = DrawPath(dc, wtext, 1);

        //Bitmap to contexts
        HBITMAP bmp = CreateCompatibleBitmap(dc, metric.x+SCALE_FACTOR, metric.y+SCALE_FACTOR);
        SelectObject(dc, bmp);

        //Draw pixels
        SetPolyFillMode(dc, WINDING);
        FillPath(dc);

        //Return pixels
        PixelsToLua(L, dc, metric, SCALE_FACTOR);

        //Free memory
        DeleteObject(dc);
        DeleteObject(bmp);
        delete (wtext);

        return 1;
    }else
        l_error(L);
    return 0;
}
