#include "generator.h"
#include <cmath>

LRESULT CALLBACK WindowProcedure (HWND, UINT, WPARAM, LPARAM);
void generate();
HANDLE thread;
GUI *wnd;
FileManager *fm;

int WINAPI WinMain (HINSTANCE hThisInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpszArgument,
                    int nCmdShow)
{
    try
    {
        wnd = new GUI(hThisInstance, WindowProcedure);
        fm = new FileManager();
        if (!fm->kt.empty())
            wnd->set_edit((wchar_t*)fm->kt.c_str() , NULL, NULL);
        if (!fm->fx.empty())
            wnd->set_edit(NULL, (wchar_t*)fm->fx.c_str(), NULL);
        if (!fm->final.empty())
            wnd->set_edit(NULL, NULL, (wchar_t*)fm->final.c_str());
    }
    catch(const char *msg)
    {
        MessageBoxA(0, msg, "Error", MB_OK | MB_ICONERROR);
        return 0;
    }

    MSG messages;
    while (GetMessage (&messages, NULL, 0, 0))
    {
        TranslateMessage(&messages);
        DispatchMessage(&messages);
    }

    delete (wnd);
    delete (fm);
    return messages.wParam;
}

LRESULT CALLBACK WindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
    case WM_CLOSE:
        if (!wnd->active())
        {
            if (IDYES == MessageBox(NULL, "Abort process?", "Abort", MB_YESNO | MB_ICONWARNING))
                DestroyWindow(hwnd);
        }
        else
            DestroyWindow(hwnd);
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    case WM_PAINT:
        wnd->paint();
        break;
    case WM_COMMAND:
        wchar_t *file;
        switch(LOWORD(wParam))
        {
        case ID_CLOSE:
            if (!wnd->active())
            {
                if (IDYES == MessageBox(NULL, "Abort process?", "Abort", MB_YESNO | MB_ICONWARNING))
                    DestroyWindow(hwnd);
            }
            else
                DestroyWindow(hwnd);
            break;
        case ID_RESET:
            wnd->reset();
            break;
        case ID_CMDS:
            ShellExecuteW(NULL, L"open", L"docs\\functions.doc", NULL, NULL, SW_MAXIMIZE);
            break;
        case ID_VALS:
            ShellExecuteW(NULL, L"open", L"docs\\ASS_values.doc", NULL, NULL, SW_MAXIMIZE);
            break;
        case ID_LUA:
            ShellExecuteW(NULL, L"open", L"docs\\lua.htm", NULL, NULL, SW_MAXIMIZE);
            break;
        case ID_ABOUT:
            MessageBoxA(hwnd,
                        "Name:    \tNyuFX\n\
Version: \t1.4.2\n\
Author:  \tYouka\n\
Website: \tforum.youka.de",
                        "About", MB_OK | MB_ICONINFORMATION);
            break;
        case ID_KTIME_BUTTON:
            if ( (file = fm->get_file_dialog(wnd, L"Input: ASS file", L"ASS file (*.ass)\0*.ass\0", L"ass")) != NULL );
            {
                wnd->set_edit(file, NULL, NULL);
                delete (file);
            }
            break;
        case ID_FX_BUTTON:
            if ( (file = fm->get_file_dialog(wnd, L"Input: LUA file", L"LUA file (*.lua)\0*.lua\0", L"lua")) != NULL );
            {
                wnd->set_edit(NULL, file, NULL);
                delete (file);
            }
            break;
        case ID_FINAL_BUTTON:
            if ( (file = fm->set_file_dialog(wnd, L"Output: ASS file", L"ASS file (*.ass)\0*.ass\0", L"ass")) != NULL );
            {
                wnd->set_edit(NULL, NULL, file);
                delete (file);
            }
            break;
        case ID_EXECUTE_BUTTON:
            if (wnd->active())
                thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&generate, 0, 0, 0);
            else
            {
                TerminateThread(thread, 0);
                wnd->enable(true);
                MessageBeep(MB_ICONSTOP);
            }
            break;
        }
        break;
    default:
        return DefWindowProc (hwnd, message, wParam, lParam);
    }
    return 0;
}

void generate()
{
    wnd->enable(false);

    //Get file names
    wchar_t **fields = wnd->get_edit();
    fm->kt = fields[0]; fm->fx = fields[1]; fm->final = fields[2];
    delete (fields[0]); delete (fields[1]); delete (fields[2]); delete (fields);

    //Check file names
    if(
        fm->fx.empty() || fm->fx.length() < 5 || fm->fx.substr(fm->fx.length()-4) != L".lua" ||
        fm->final.empty() || fm->final.length() < 5 || fm->final.substr(fm->final.length()-4) != L".ass" ||
        ( !fm->kt.empty() && (fm->kt.length() < 5 || fm->kt.substr(fm->kt.length()-4) != L".ass") )
       )
        MessageBoxW(NULL, L"Invalid file names!", L"Wrong input", MB_OK | MB_ICONERROR);
    else
    {
        //Try to generate
        Generator gen(fm, wnd);
        if( gen.run() )
        {
            unsigned int t = gen.get_time();
            unsigned int l = gen.get_lines();

            wchar_t *wtext = new wchar_t[17 + int(log10(l))+1];
            swprintf(wtext, L"Produced lines: %d", l);
            wnd->log_append(wtext);
            delete (wtext);

            wtext = new wchar_t[33 + int(log10(l))+1+4];
            swprintf(wtext, L"Progress duration (in seconds): %.3f", double(t) / 1000);
            wnd->log_append(wtext);
            delete (wtext);

            MessageBeep(MB_ICONINFORMATION);
        }
    }

    wnd->enable(true);
}
