function sub(line)
	local l = table.copy(line)

	--basic
	l.text = string.format("{\\pos(%.3f,%.3f)\\blur3}%s", line.x, line.y, line.text)
	Output(l)

	--infade
	for s, e, i, n in frames(line.start_time - line.infade/2, line.start_time, 41.71) do
		l.start_time = s
		l.end_time = e
		local frame_pct = i / n
		local alpha = A_to_ASS( frame_pct * 255 )
		local line_path = GetPath(line.text, line.styleref)
		line_path:split(3)
		local function wobble_filter(point)
			local x_add = math.sin(point[2] / line.height * math.pi / 2 + frame_pct * math.pi * 4) * 40
			local y_add = math.cos(point[1] / line.width * math.pi * 4 + frame_pct * math.pi * 4) * 40
			point[1] = point[1] + x_add * (1-frame_pct)
			point[2] = point[2] + y_add * (1-frame_pct)
		end
		line_path:filter(wobble_filter)
		l.text = string.format("{\\blur3\\pos(%.3f,%.3f)\\alpha%s\\an7\\p4}%s", line.left, line.top, alpha, line_path:tostring())
		Output(l)
	end

	--outfade
	for s, e, i, n in frames(line.end_time, line.end_time + line.outfade/2, 41.71) do
		l.start_time = s
		l.end_time = e
		local frame_pct = i / n
		local alpha = A_to_ASS( (1 - frame_pct) * 255 )
		local line_path = GetPath(line.text, line.styleref)
		line_path:split(3)
		local function wobble_filter(point)
			local x_add = math.sin(point[2] / line.height * math.pi / 2 + frame_pct * math.pi * 4) * 40
			local y_add = math.cos(point[1] / line.width * math.pi * 4 + frame_pct * math.pi * 4) * 40
			point[1] = point[1] + x_add * frame_pct
			point[2] = point[2] + y_add * frame_pct
		end
		line_path:filter(wobble_filter)
		l.text = string.format("{\\blur3\\pos(%.3f,%.3f)\\alpha%s\\an7\\p4}%s", line.left, line.top, alpha, line_path:tostring())
		Output(l)
	end
end

function roumaji(line)
	local l = table.copy(line)

	l.style = "p"
	local rect = Rectangle(1,1)
	l.start_time = line.start_time - line.infade/2
	for ci, char in ipairs(line.chars) do
		local fx_duration = char.duration + 300
		l.end_time = line.start_time + char.start_time + fx_duration
		for pi, pixel in ipairs( GetPath(char.text, line.styleref):topixels() ) do
			local x1, y1 = char.left + pixel.x, char.top + pixel.y
			local x2, y2 = x1 + math.random(-20, 20), y1 + math.random(-20, 20)
			local alpha = (pixel.a == 255 and "") or ("\\alpha" .. A_to_ASS(pixel.a))
			local color = Interpolate(pixel.y / char.height, "&H0000FF&", "&H00FFFF&", "curve_up")
			l.text = string.format("{\\fad(%d,%d)\\1c%s%s\\move(%.3f,%.3f,%.3f,%.3f,%d,%d)\\p1}%s",
									line.infade/2, fx_duration, color, alpha, x1, y1, x2, y2, line.infade/2 + char.start_time, line.infade/2 + char.start_time + fx_duration, rect)
			Output(l)
		end
	end
end

function kanji(line)
	local l = table.copy(line)

	l.style = "p"
	local rect = Rectangle(1,1)
	l.start_time = line.start_time - line.infade/2
	for ci, char in ipairs(line.chars) do
		local fx_duration = char.duration + 300
		l.end_time = line.start_time + char.start_time + fx_duration
		for pi, pixel in ipairs( GetPath(char.text, line.styleref):topixels() ) do
			local x1, y1 = char.left + pixel.x, char.top + pixel.y
			local x2, y2 = x1 + math.random(-20, 20), y1 + math.random(-20, 20)
			local alpha = (pixel.a == 255 and "") or ("\\alpha" .. A_to_ASS(pixel.a))
			local color = Interpolate(pixel.x / char.width, "&H0000FF&", "&H00FFFF&", "curve_up")
			l.text = string.format("{\\fad(%d,%d)\\1c%s%s\\move(%.3f,%.3f,%.3f,%.3f,%d,%d)\\p1}%s",
									line.infade/2, fx_duration, color, alpha, x1, y1, x2, y2, line.infade/2 + char.start_time, line.infade/2 + char.start_time + fx_duration, rect)
			Output(l)
		end
	end
end

function main()
	for li, line in ipairs(lines) do
		if line.styleref.alignment >=7 then
			roumaji(line)
		elseif line.styleref.alignment <=3 then
			sub(line)
		else
			kanji(line)
		end
		ProgressStatus(li / #lines)
	end
	print("Writing output file...")
end

main()
