function FadePoints_to_Pixels(pixel_table, fade_points_distance)
	--Width and height
	local width, height = 0, 0
	for pixel_i, pixel in ipairs(pixel_table) do
		width = math.max(width, pixel.x)
		height = math.max(height, pixel.y)
	end
	--Fade points
	local fade_points = {}
	local fade_points_offset = math.mod(width, fade_points_distance) / 2
	local fade_points_n = (width - fade_points_offset * 2) / fade_points_distance + 1
	for fade_i = 1, fade_points_n do
		fade_points[fade_i] = {
								x = fade_points_offset + (fade_i-1) * fade_points_distance + math.random(-fade_points_distance / 4, fade_points_distance / 4),
								y = height / 2 + math.random(-height / 2, height / 2)
							}
	end
	--Pixel distance to nearest point
	for pixel_i, pixel in ipairs(pixel_table) do
		for fade_i, fade_point in ipairs(fade_points) do
			local cur_distance = math.distance(pixel.x - fade_point.x, pixel.y - fade_point.y)
			pixel.distance = fade_i == 1 and cur_distance or math.min(pixel.distance, cur_distance)
		end
	end
	--Search maximal distance
	local max_distance = 0
	for pixel_i, pixel in ipairs(pixel_table) do
		max_distance = math.max(max_distance, pixel.distance)
	end
	--Pixels distance in percent
	for pixel_i, pixel in ipairs(pixel_table) do
		pixel.pct = pixel.distance / max_distance
		pixel.distance = nil
	end
end

function sub(line)
	local l = table.copy(line)

	l.style = "p"
	local rect = Rectangle(1,1)
	local pixels = GetPath(line.text, line.styleref):topixels()
	FadePoints_to_Pixels(pixels, 50)
	for pi, pixel in ipairs(pixels) do
		local x, y = line.left + pixel.x, line.top + pixel.y
		local alpha = (pixel.a == 255 and "") or ("\\alpha" .. A_to_ASS(pixel.a))
		local color = Interpolate(pixel.pct, "&HFFFFFF&", "&H000000&")
		local fad_in, fad_out
		if line.infade/2 >= 126 then
			fad_in = 126
			l.start_time = line.start_time - line.infade/2 + pixel.pct * (line.infade/2 - 126)
		else
			fad_in = 0
			l.start_time = line.start_time
		end
		if line.outfade/2 >= 126 then
			fad_out = 126
			l.end_time = line.end_time + pixel.pct * (line.outfade/2 - 126)
		else
			fad_out = 0
			l.end_time = line.end_time
		end
		l.text = string.format("{\\fad(%d,%d)\\1c%s%s\\pos(%.3f,%.3f)\\p1}%s", fad_in, fad_out, color, alpha, x, y, rect)
		Output(l)
	end
end

function main()
	for li, line in ipairs(lines) do
		if line.styleref.alignment <=3 then
			sub(line)
		end
		ProgressStatus(li / #lines)
	end
	print("Writing output file...")
end

main()
