function sub(line)
	local l = table.copy(line)
	l.start_time = line.start_time - line.infade/2
	l.end_time = line.end_time + line.outfade/2
	l.text = string.format("{\\fad(%d,%d)}%s", line.infade/2, line.outfade/2, line.text)
	Output(l)
end

function roumaji_chor(line)
	local l = table.copy(line)
	for si, syl in ipairs(line.syls) do
		for s, e, i, n in frames(line.start_time + syl.start_time, line.start_time + syl.end_time, 41.71*8) do
			l.start_time = s
			l.end_time = s + 500
			local x = line.center + 2*(syl.x - line.center)
			l.text = string.format("{\\an5\\move(%.3f,%.3f,%.3f,%.3f)\\t(\\fscx200\\fscy200)\\fad(%d,%d)}%s"
									, x, syl.middle, x, syl.middle-syl.height/8, 80, 420, syl.text)
			Output(l)
		end
	end
end

--Clef, lines and note shapes
local clef =
"m 15 32 b 19 0 29 14 29 28 b 30 47 8 52 6 68 b 8 89 34 86 34 72 b 35 60 12 56 14 74 b 15 78 19 79 21 80 b 4 78 13 56 21 57 b 44 57 43 84 21 85 b 12 84 0 78 2 65 b 7 41 26 44 27 25 b 27 16 17 15 18 35 l 29 97 b 28 106 17 112 11 101 b 8 85 32 96 16 104 b 23 109 27 100 28 97 "
local note = "m 15 20 l 15 2 b 15 0 17 1 17 2 b 20 8 29 12 26 21 b 25 23 24 22 24 18 b 24 14 19 7 17 10 l 17 33 b 13 42 1 41 1 36 b 0 30 7 27 12 28 b 14 30 15 28 15 26 l 15 20"
local line_len, line_width = 10, 2
local line_rows = ""
for i=0, 4 do
	local y = i * (lines[1].height/4)
	line_rows = line_rows .. string.format("m 0 %d l %d %d %d %d 0 %d ", y, line_len, y, line_len, y+line_width, y+line_width)
end
local margin = 70

function roumaji_sing(line)
	local l = table.copy(line)
	--Clef & lines infade & outfade
	l.start_time = line.start_time - line.infade
	local clef_infade, clef_outfade
	local line_infade, line_outfade
	local fscx = (line.width + margin)/line_len * 100
	if math.mod(line.infade, 1) ~= 0 then
		clef_infade = string.format("\\fad(%d,0)\\pos(%.3f,%.3f)", line.infade, line.left-margin/2, line.middle)
		line_infade = string.format("\\pos(%.3f,%.3f)\\fad(%d,0)\\fscx%d", line.x, line.y, line.infade, fscx)
	elseif line.infade == 0 then
		clef_infade = string.format("\\pos(%.3f,%.3f)", line.left-margin/2, line.middle)
		line_infade = string.format("\\pos(%.3f,%.3f)\\fscx%d", line.x, line.y, fscx)
	else
		clef_infade = string.format("\\move(%.3f,%.3f,%.3f,%.3f,0,%d)", lines[line.i-1].left-margin/2, lines[line.i-1].middle, line.left-margin/2, line.middle, line.infade)
		local pre_fscx = (lines[line.i-1].width+margin)/line_len * 100
		line_infade = string.format("\\pos(%.3f,%.3f)\\fscx%d\\t(0,%d,\\fscx%d)", line.x, line.y, pre_fscx, line.infade, fscx)
	end
	if math.mod(line.outfade, 1) ~= 0 then
		l.end_time = line.end_time + line.outfade
		clef_outfade = string.format("\\fad(0,%d)", line.outfade)
		line_outfade = string.format("\\fad(0,%d)", line.outfade)
	else
		clef_outfade = ""
		line_outfade = ""
	end
	--Lines
	l.text = string.format("{%s%s\\1c&HB0B0B0&\\bord0\\p1}%s", line_infade, line_outfade, line_rows)
	Output(l)
	--Clef
	l.text = string.format("{%s%s\\1c&H000000&\\an4\\bord0\\fscx70\\fscy70\\p1}%s", clef_infade, clef_outfade, clef)
	Output(l)
	--Jumping note
	local x_fix, y_fix = 2, 2
	local in_out_way = 30
	local jump_height = 5
	for si, syl in ipairs(line.syls) do
		local _, min_y = GetPath(syl.text, line.styleref):getbound()
		local x_start, y_start = syl.center + x_fix, syl.top + min_y + y_fix
		local x_way, y_way
		if si == #line.syls then
			x_way = in_out_way
			y_way = 0
		else
			x_way = line.syls[si+1].center - x_start
			_, min_y = GetPath(line.syls[si+1].text, line.styleref):getbound()
			y_way = (line.syls[si+1].top + min_y) - y_start
		end
		for s, e, i, n in frames(line.start_time+syl.start_time, line.start_time+syl.end_time, 41.71) do
			l.start_time = s
			l.end_time = e
			local frame_pct = i/n
			local x = x_start + frame_pct * x_way
			local y = y_start + frame_pct * y_way - math.sin(frame_pct * math.pi) * jump_height
			l.text = string.format("{\\pos(%.3f,%.3f)\\an2\\1c&H000000&\\bord0\\p2}%s", x, y, note)
			if si == #line.syls then
				local alpha = A_to_ASS(255 - 255 * frame_pct)
				l.text = l.text:gsub("{", "{\\alpha"..alpha)
			end
			Output(l)
		end
		if si == 1 then
			for s, e, i, n in frames(line.start_time - 200, line.start_time, 41.71) do
				l.start_time = s
				l.end_time = e
				local frame_pct = i / n
				local x = x_start - in_out_way + frame_pct * in_out_way
				local y = y_start - math.sin(math.pi / 2 + frame_pct * math.pi / 2) * jump_height
				local alpha = A_to_ASS(255 * frame_pct)
				l.text = string.format("{\\pos(%.3f,%.3f)\\alpha%s\\an2\\1c&H000000&\\bord0\\p2}%s", x, y, alpha, note)
				Output(l)
			end
		end
	end
	--Basic text
	l.layer = 1
	l.start_time = line.start_time
	for si, syl in ipairs(line.syls) do
		l.end_time = line.start_time + syl.start_time
		l.text = string.format("{\\pos(%.3f,%.3f)}%s", syl.x, syl.y, syl.text)
		Output(l)
	end
	l.end_time = line.end_time
	for si, syl in ipairs(line.syls) do
		l.start_time = line.start_time + syl.end_time
		l.text = string.format("{\\pos(%.3f,%.3f)}%s", syl.x, syl.y, syl.text)
		Output(l)
	end
	--Sylables effect
	local wave_range = 10
	for si, syl in ipairs(line.syls) do
		if syl.text ~= "" then
			local syl_path = GetPath(syl.text, line.styleref)
			local _, min_y, _, _, max_y = syl_path:getbound()
			local dist_y = max_y - min_y
			for s, e, i, n in frames(line.start_time+syl.start_time, line.start_time+syl.end_time, 41.71) do
				l.start_time = s
				l.end_time = e
				local frame_pct = i/n
				local y_anchor = min_y - wave_range + (dist_y + 2 * wave_range) * frame_pct
				syl_path = GetPath(syl.text, line.styleref)
				syl_path:split(3)
				local function syl_fx(p)
					local cur_dist = math.abs(p[2]/8 - y_anchor)
					if cur_dist <= wave_range then
						local pct = 1 - (cur_dist / wave_range)
						local x_add = math.sin(p[2]/15) * 40 * pct
						p[1] = p[1] + x_add
					end
				end
				syl_path:filter(syl_fx)
				l.text = string.format("{\\pos(%.3f,%.3f)\\an7\\p4}%s", syl.left, syl.top, syl_path:tostring())
				Output(l)
			end
		end
	end
	--Infade text
	l.end_time = line.start_time
	if math.mod(line.infade, 1) ~= 0 then
		l.start_time = line.start_time - line.infade
		l.text = string.format("{\\pos(%.3f,%.3f)\\fad(%d,0)}%s", line.x, line.y, line.infade, line.text)
		Output(l)
	else
		l.start_time = line.start_time - line.infade / 2
		local fscx = (lines[line.i-1].width / line.width - 1) * 100 / 2 + 100
		l.text = string.format("{\\pos(%.3f,%.3f)\\fad(%d,0)\\fscx%d\\t(\\fscx100)}%s", line.x, line.y, line.infade/2, fscx, line.text)
		Output(l)
	end
	--Outfade text
	if line.i ~= 7 then
		l.start_time = line.end_time
		if math.mod(line.outfade, 1) ~= 0 then
			l.end_time = line.end_time + line.outfade
			l.text = string.format("{\\pos(%.3f,%.3f)\\fad(0,%d)}%s", line.x, line.y, line.outfade, line.text)
			Output(l)
		else
			l.end_time = line.end_time + line.outfade / 2
			local fscx = (lines[line.i+1].width / line.width - 1) * 100 / 2 + 100
			l.text = string.format("{\\pos(%.3f,%.3f)\\fad(0,%d)\\t(\\fscx%d)}%s", line.x, line.y, line.outfade/2, fscx, line.text)
			Output(l)
		end
	end
end

function main()
	for li, line in ipairs(lines) do
		if line.styleref.alignment >=7 then
			if line.text == "aoo" then
				roumaji_chor(line)
			else
				roumaji_sing(line)
			end
		elseif line.styleref.alignment <=3 then
			sub(line)
		end
		ProgressStatus(li / #lines)
	end
	print("Writing output file...")
end

main()
